#include "mlx/backend/metal/kernels/svd.h"
#include "mlx/allocator.h"
#include "mlx/backend/common/compiled.h"
#include "mlx/backend/common/copy.h"
#include "mlx/backend/gpu/copy.h"
#include "mlx/backend/metal/device.h"
#include "mlx/backend/metal/kernels.h"
#include "mlx/backend/metal/utils.h"
#include "mlx/ops.h"
#include "mlx/primitives.h"
#include "mlx/scheduler.h"

/**
 * Implementation of a full GPU-accelerated SVD using the one-sided Jacobi
 * algorithm.
 * - Computes A^T*A and diagonalizes it using Jacobi rotations
 * - Singular values: σᵢ = √λᵢ where λᵢ are eigenvalues of A^T*A
 * - Right singular vectors: V from eigenvectors of A^T*A
 * - Left singular vectors: U = A*V*Σ^-1
 *
 * - Precision: Float32 (Metal limitation)
 */

namespace mlx::core {

namespace {

/**
 * Select appropriate SVD algorithm based on matrix properties
 */
enum class SVDAlgorithm {
  JACOBI_ONE_SIDED, // Implemented - Default for most cases
  JACOBI_TWO_SIDED, // Future: Better numerical stability for ill-conditioned
                    // matrices
  BIDIAGONAL_QR // Future: For very large matrices (>4096x4096)
};

SVDAlgorithm select_svd_algorithm(int M, int N, Dtype dtype) {
  // Algorithm selection based on matrix properties

  // For very large matrices, we might want different algorithms in the future
  if (std::max(M, N) > 2048) {
    // Currently use Jacobi for all sizes up to 4096x4096
    // Future: Could implement bidiagonal QR for better performance on large
    // matrices
    return SVDAlgorithm::JACOBI_ONE_SIDED;
  }

  // For very rectangular matrices, one-sided Jacobi is efficient
  double aspect_ratio = static_cast<double>(std::max(M, N)) / std::min(M, N);
  if (aspect_ratio > 3.0) {
    return SVDAlgorithm::JACOBI_ONE_SIDED;
  }

  // Default to one-sided Jacobi for most cases
  return SVDAlgorithm::JACOBI_ONE_SIDED;
}

/**
 * Compute SVD parameters based on matrix size and algorithm
 */
SVDParams compute_svd_params(
    int M,
    int N,
    size_t num_matrices,
    bool compute_uv,
    SVDAlgorithm algorithm) {
  const int K = std::min(M, N);

  // Adjust parameters based on matrix size and algorithm
  int max_iterations = 100;
  float tolerance = 1e-6f;

  // For larger matrices, we might need more iterations
  if (std::max(M, N) > 512) {
    max_iterations = 200;
    tolerance = 1e-5f; // Slightly relaxed tolerance for large matrices
  }

  // For very small matrices, we can use tighter tolerance
  if (std::max(M, N) < 64) {
    tolerance = 1e-7f;
  }

  return SVDParams{
      M, // M
      N, // N
      K, // K
      max_iterations, // max_iterations
      tolerance, // tolerance
      static_cast<int>(num_matrices), // batch_size
      M * N, // matrix_stride
      compute_uv // compute_uv
  };
}

/**
 * Validate SVD input parameters
 */
void validate_svd_inputs(const array& a) {
  if (a.ndim() < 2) {
    throw std::invalid_argument(
        "[SVD::eval_gpu] Input must have >= 2 dimensions, got " +
        std::to_string(a.ndim()) + "D array");
  }

  if (a.dtype() != float32 && a.dtype() != float64) {
    throw std::invalid_argument(
        "[SVD::eval_gpu] Only float32 and float64 supported, got " +
        type_to_name(a.dtype()));
  }

  // Note: Metal does not support double precision, will fall back to CPU
  if (a.dtype() == float64) {
    throw std::runtime_error(
        "[SVD::eval_gpu] Double precision not supported on Metal GPU. "
        "Use mx.set_default_device(mx.cpu) for float64 SVD operations.");
  }

  // Check for reasonable matrix size
  int M = a.shape(-2);
  int N = a.shape(-1);
  if (M > 4096 || N > 4096) {
    throw std::invalid_argument(
        "[SVD::eval_gpu] Matrix too large for current implementation. "
        "Got " +
        std::to_string(M) + "x" + std::to_string(N) +
        ", maximum supported size is 4096x4096");
  }

  if (M == 0 || N == 0) {
    throw std::invalid_argument(
        "[SVD::eval_gpu] Matrix dimensions must be positive, got " +
        std::to_string(M) + "x" + std::to_string(N));
  }

  // Check for empty arrays
  if (a.size() == 0) {
    throw std::invalid_argument("[SVD::eval_gpu] Input matrix is empty");
  }

  // Note: Input validation is performed here rather than during evaluation
  // to avoid recursive evaluation issues with Metal command buffers
}

} // anonymous namespace

template <typename T>
void svd_metal_impl(
    const array& a,
    std::vector<array>& outputs,
    bool compute_uv,
    metal::Device& d,
    const Stream& s) {
  // Validate inputs
  validate_svd_inputs(a);

  // Matrix dimensions
  const int M = a.shape(-2);
  const int N = a.shape(-1);
  const int K = std::min(M, N);
  const size_t batch_size = a.size() / (M * N);

  // SVD parameters
  SVDParams params = {
      .M = M,
      .N = N,
      .K = K,
      .max_iterations = 100, // Maximum Jacobi iterations
      .tolerance = 1e-6f, // Convergence threshold
      .batch_size = static_cast<int>(batch_size),
      .matrix_stride = M * N,
      .compute_uv = compute_uv};

  // Allocate memory for all outputs
  for (auto& output : outputs) {
    if (output.size() > 0) {
      output.set_data(allocator::malloc(output.nbytes()));
    }
  }

  // Get Metal command encoder (MLX manages the command buffer lifecycle)
  auto& compute_encoder = d.get_command_encoder(s.index);

  // Use a SINGLE comprehensive kernel that performs the entire SVD computation
  // This follows MLX patterns where each primitive dispatches only one kernel
  auto kernel = d.get_kernel("svd_jacobi_complete_float");
  compute_encoder.set_compute_pipeline_state(kernel);

  // Set input and output arrays
  compute_encoder.set_input_array(a, 0);
  if (compute_uv) {
    compute_encoder.set_output_array(outputs[0], 1); // U
    compute_encoder.set_output_array(outputs[1], 2); // S
    compute_encoder.set_output_array(outputs[2], 3); // Vt
  } else {
    compute_encoder.set_output_array(outputs[0], 1); // S only
  }

  // Set parameters
  compute_encoder.set_bytes(&params, sizeof(SVDParams), 4);

  // Dispatch the comprehensive kernel
  // Use a grid that can handle the entire computation
  MTL::Size grid_size = MTL::Size(std::max(M, N), std::max(M, N), batch_size);
  MTL::Size group_size = MTL::Size(16, 16, 1);
  compute_encoder.dispatch_threads(grid_size, group_size);

  // MLX automatically handles command buffer commit and completion handlers
  // No manual command buffer management needed
}

// Explicit template instantiation for float32 only
// Note: Metal does not support double precision
template void svd_metal_impl<float>(
    const array& a,
    std::vector<array>& outputs,
    bool compute_uv,
    metal::Device& d,
    const Stream& s);

} // namespace mlx::core
